@@ -1,5 +1,5 @@
 // SPDX-License-Identifier: LGPL-2.1-only
-// Copyright (c) 2015 - 2020 DisplayLink (UK) Ltd.
+// Copyright (c) 2015 - 2024 DisplayLink (UK) Ltd.
 
 #include <stddef.h>
 #include <stdint.h>
@@ -23,13 +23,14 @@
 #include <sys/types.h>
 #include <unistd.h>
 
+#define MIN(x, y) (((x) < (y)) ? (x) : (y))
 // ********************* Private part **************************
 
 #define MAX_DIRTS           16
 #define EVDI_INVALID_DEVICE_INDEX -1
 
 #define EVDI_MODULE_COMPATIBILITY_VERSION_MAJOR 1
-#define EVDI_MODULE_COMPATIBILITY_VERSION_MINOR 8
+#define EVDI_MODULE_COMPATIBILITY_VERSION_MINOR 9
 #define EVDI_MODULE_COMPATIBILITY_VERSION_PATCH 0
 
 #define evdi_log(...) do {						\
@@ -130,11 +131,9 @@ static struct evdi_frame_buffer_node *findBuffer(evdi_handle context, int id)
 static void addFrameBuffer(evdi_handle context,
 			   struct evdi_buffer const *frame_buffer)
 {
-	struct evdi_frame_buffer_node **node = NULL;
-
-	for (node = &context->frameBuffersListHead;
+	for (struct evdi_frame_buffer_node **node = &context->frameBuffersListHead;
 	     ;
-	     node = (struct evdi_frame_buffer_node **)&(*node)->next) {
+	     node = &(*node)->next) {
 		if (*node)
 			continue;
 
@@ -149,15 +148,12 @@ static void addFrameBuffer(evdi_handle context,
  * @brief Removes all frame buffers matching the given id
  * @param id of frame buffer to remove, NULL matches every buffer, thus all
  * will be removed
- * @return number of buffers removed
- * @todo Return value doesn't seem to be used anywhere
  */
-static int removeFrameBuffer(evdi_handle context, int const *id)
+static void removeFrameBuffer(evdi_handle context, int const *id)
 {
 	struct evdi_frame_buffer_node *current = NULL;
 	struct evdi_frame_buffer_node *next = NULL;
 	struct evdi_frame_buffer_node **prev = NULL;
-	int removedCount = 0;
 
 	current = context->frameBuffersListHead;
 	prev = &context->frameBuffersListHead;
@@ -166,7 +162,6 @@ static int removeFrameBuffer(evdi_handle context, int const *id)
 
 		if (!id || current->frame_buffer.id == *id) {
 			free(current);
-			++removedCount;
 			*prev = next;
 		} else {
 			prev = &current->next;
@@ -174,8 +169,6 @@ static int removeFrameBuffer(evdi_handle context, int const *id)
 
 		current = next;
 	}
-
-	return removedCount;
 }
 
 static int is_evdi_compatible(int fd)
@@ -209,7 +202,9 @@ static int is_evdi_compatible(int fd)
 
 static int is_evdi(int fd)
 {
-	char name[64] = { 0 }, date[64] = { 0 }, desc[64] = { 0 };
+	char name[64] = { 0 };
+	char date[64] = { 0 };
+	char desc[64] = { 0 };
 	struct drm_version ver = {
 		.name_len = sizeof(name),
 		.name = name,
@@ -248,6 +243,7 @@ static int does_path_links_to(const char *link, const char *substr)
 	r = readlink(link, real_path, sizeof(real_path));
 	if (r < 0)
 		return 0;
+	r = MIN(r, PATH_MAX-1);
 	real_path[r] = '\0';
 
 	return (strstr(real_path, substr) != NULL);
@@ -310,7 +306,7 @@ static int device_has_master(const char *device_file_path)
 {
 	pid_t myself = getpid();
 	DIR *proc_dir;
-	struct dirent *proc_entry;
+	const struct dirent *proc_entry;
 	int result = 0;
 
 	proc_dir = opendir("/proc");
@@ -318,7 +314,7 @@ static int device_has_master(const char *device_file_path)
 		return 0;
 
 	while ((proc_entry = readdir(proc_dir)) != NULL) {
-		char *d_name = proc_entry->d_name;
+		const char *d_name = proc_entry->d_name;
 
 		if (d_name[0] < '0'
 		    || d_name[0] > '9'
@@ -414,9 +410,8 @@ static int open_device(int device)
 
 	snprintf(dev, PATH_MAX, "/dev/dri/card%d", device);
 
-#ifndef CHROMEOS
-	wait_for_master(dev);
-#endif
+	if (Xorg_running())
+		wait_for_master(dev);
 
 	fd = wait_for_device(dev);
 
@@ -468,15 +463,43 @@ static int get_drm_device_index(const char *evdi_sysfs_drm_dir)
 	return dev_index;
 }
 
-static int find_unused_card_for(const char *sysfs_parent_device_path)
+static bool is_correct_parent_device(const char *dirname, size_t dirname_maxlen, const char *parent_device, size_t parent_length)
 {
+	char link_path[PATH_MAX];
+
+	snprintf(link_path, MIN(PATH_MAX - 7, dirname_maxlen), "%s/device", dirname);
+
+	if (parent_device == NULL)
+		return access(link_path, F_OK) != 0;
+
+	char link_resolution[PATH_MAX];
+	const ssize_t link_resolution_len = readlink(link_path, link_resolution, PATH_MAX);
+
+	if (link_resolution_len < 0 || link_resolution_len >= PATH_MAX)
+		return false;
+
+	link_resolution[link_resolution_len] = '\0';
+	char *parent_device_token = strrchr(link_resolution, '/');
+
+	if (parent_length < 2)
+		return false;
+
+	parent_device_token++;
+	size_t len = strnlen(parent_device_token, parent_length - (parent_device_token - parent_device));
+	bool is_same_device = parent_length == len && strncmp(parent_device_token, parent_device, len) == 0;
+	return is_same_device;
+}
+
+static int find_unused_card_for(const char *parent_device, size_t length)
+{
+	char evdi_platform_root[] = "/sys/bus/platform/devices";
 	struct dirent *fd_entry;
 	DIR *fd_dir;
 	int device_index = EVDI_INVALID_DEVICE_INDEX;
 
-	fd_dir = opendir(sysfs_parent_device_path);
+	fd_dir = opendir(evdi_platform_root);
 	if (fd_dir == NULL) {
-		evdi_log("Failed to open dir %s", sysfs_parent_device_path);
+		evdi_log("Failed to open dir %s", evdi_platform_root);
 		return device_index;
 	}
 
@@ -484,9 +507,16 @@ static int find_unused_card_for(const char *sysfs_parent_device_path)
 		if (strncmp(fd_entry->d_name, "evdi", 4) != 0)
 			continue;
 
+		char evdi_path[PATH_MAX];
+
+		snprintf(evdi_path, PATH_MAX, "%s/%s", evdi_platform_root, fd_entry->d_name);
+
+		if (!is_correct_parent_device(evdi_path, PATH_MAX,  parent_device, length))
+			continue;
+
 		char evdi_drm_path[PATH_MAX];
 
-		snprintf(evdi_drm_path, PATH_MAX, "%s/%s/drm", sysfs_parent_device_path, fd_entry->d_name);
+		snprintf(evdi_drm_path, PATH_MAX - strnlen(evdi_path, PATH_MAX), "%s/drm", evdi_path);
 		int dev_index = get_drm_device_index(evdi_drm_path);
 
 		assert(dev_index < EVDI_USAGE_LEN && dev_index >= 0);
@@ -503,37 +533,34 @@ static int find_unused_card_for(const char *sysfs_parent_device_path)
 
 static int get_generic_device(void)
 {
-	char evdi_platform_root[] = "/sys/devices/platform";
 	int device_index = EVDI_INVALID_DEVICE_INDEX;
 
-	device_index = find_unused_card_for(evdi_platform_root);
+	device_index = find_unused_card_for(NULL, 0);
 	if (device_index == EVDI_INVALID_DEVICE_INDEX) {
-		evdi_log("Creating card in %s", evdi_platform_root);
+		evdi_log("Creating card in /sys/devices/platform");
 		write_add_device("1", 1);
-		device_index = find_unused_card_for(evdi_platform_root);
+		device_index = find_unused_card_for(NULL, 0);
 	}
 
 	return device_index;
 }
 
-static int get_device_attached_to_usb(const char *bus_ident)
+static int get_device_attached_to_usb(const char *sysfs_parent_device, size_t length)
 {
+	const size_t USB_OFFSET = 4;
 	int device_index = EVDI_INVALID_DEVICE_INDEX;
-	char evdi_usb_parent_path[PATH_MAX] = "/sys/bus/usb/devices/";
 
-	strncat(evdi_usb_parent_path, bus_ident, PATH_MAX - strlen(evdi_usb_parent_path));
+	if (length < USB_OFFSET)
+		return EVDI_INVALID_DEVICE_INDEX;
+	const char *parent_device = &sysfs_parent_device[USB_OFFSET];
+	const size_t parent_length = length - USB_OFFSET;
 
-	device_index = find_unused_card_for(evdi_usb_parent_path);
+	device_index = find_unused_card_for(parent_device, parent_length);
 	if (device_index == EVDI_INVALID_DEVICE_INDEX) {
-		evdi_log("Creating card for %s", bus_ident);
-		char usb_dev_path[PATH_MAX] = "usb:";
-		const size_t current_len = strlen(usb_dev_path);
+		evdi_log("Creating card for usb device %s in /sys/bus/platform/devices", parent_device);
 
-		strncat(usb_dev_path, bus_ident, PATH_MAX - current_len);
-		const size_t len = strlen(usb_dev_path);
-
-		write_add_device(usb_dev_path, len);
-		device_index = find_unused_card_for(evdi_usb_parent_path);
+		write_add_device(sysfs_parent_device, length);
+		device_index = find_unused_card_for(parent_device, parent_length);
 	}
 
 	return device_index;
@@ -609,21 +636,21 @@ int evdi_add_device(void)
 	return write_add_device("1", 1);
 }
 
+// deprecated, use evdi_open_attached_to_fixed
 evdi_handle evdi_open_attached_to(const char *sysfs_parent_device)
+{
+	return evdi_open_attached_to_fixed(sysfs_parent_device, strlen(sysfs_parent_device));
+}
+
+evdi_handle evdi_open_attached_to_fixed(const char *sysfs_parent_device, size_t length)
 {
 	int device_index = EVDI_INVALID_DEVICE_INDEX;
 
 	if (sysfs_parent_device == NULL)
 		device_index = get_generic_device();
 
-	if (sysfs_parent_device != NULL && strncmp(sysfs_parent_device, "usb:", 4) == 0 && strlen(sysfs_parent_device) > 4) {
-		char bus_ident[PATH_MAX];
-		const size_t len = strlen(sysfs_parent_device) - 4;
-
-		strncpy(bus_ident, &sysfs_parent_device[4], len);
-		bus_ident[len] = 0;
-		device_index = get_device_attached_to_usb(bus_ident);
-	}
+	if (sysfs_parent_device != NULL && length > 4 && strncmp(sysfs_parent_device, "usb:", 4) == 0)
+		device_index = get_device_attached_to_usb(sysfs_parent_device, length);
 
 	if (device_index >= 0 && device_index < EVDI_USAGE_LEN)  {
 		evdi_handle handle = evdi_open(device_index);
@@ -652,13 +679,23 @@ void evdi_connect(evdi_handle handle,
 		  const unsigned char *edid,
 		  const unsigned int edid_length,
 		  const uint32_t sku_area_limit)
+{
+	evdi_connect2(handle, edid, edid_length, sku_area_limit, sku_area_limit*60);
+}
+
+void evdi_connect2(evdi_handle handle,
+		  const unsigned char *edid,
+		  const unsigned int edid_length,
+		  const uint32_t pixel_area_limit,
+		  const uint32_t pixel_per_second_limit)
 {
 	struct drm_evdi_connect cmd = {
 		.connected = 1,
 		.dev_index = handle->device_index,
 		.edid = edid,
 		.edid_length = edid_length,
-		.sku_area_limit = sku_area_limit,
+		.pixel_area_limit = pixel_area_limit,
+		.pixel_per_second_limit = pixel_per_second_limit,
 	};
 
 	do_ioctl(handle->fd, DRM_IOCTL_EVDI_CONNECT, &cmd, "connect");
@@ -666,44 +703,22 @@ void evdi_connect(evdi_handle handle,
 
 void evdi_disconnect(evdi_handle handle)
 {
-	struct drm_evdi_connect cmd = { 0, 0, 0, 0, 0 };
+	struct drm_evdi_connect cmd = { 0, 0, 0, 0, 0, 0 };
 
 	do_ioctl(handle->fd, DRM_IOCTL_EVDI_CONNECT, &cmd, "disconnect");
 }
 
-void evdi_enable_cursor_events(evdi_handle handle)
+void evdi_enable_cursor_events(evdi_handle handle, bool enable)
 {
-	char path[PATH_MAX] = {0};
-	static const char enable[] = "Y";
-	int path_len = 0;
-	FILE *cursor_evs = NULL;
-	size_t written = 0;
-	const size_t elem_bytes = 1;
-	int errcode = 0;
+	struct drm_evdi_enable_cursor_events cmd = {
+		.enable = enable,
+	};
 
-	if (evdi_device_to_platform(handle->device_index, path) !=
-	    AVAILABLE) {
-		evdi_log("Failed to enable cursor events");
-		evdi_log("Device /dev/dri/card%d, device is not available.",
-			handle->device_index);
-		return;
-	}
+	evdi_log("%s cursor events on /dev/dri/card%d",
+		(enable ? "Enabling" : "Disabling"),
+		handle->device_index);
 
-	path_len = strlen(path);
-	snprintf(path+path_len, PATH_MAX-path_len, "/cursor_events");
-	cursor_evs = fopen(path, "w");
-	if (cursor_evs == NULL) {
-		evdi_log("Failed to open %s, err: %s", path, strerror(errno));
-		return;
-	}
-
-	written = fwrite(enable, elem_bytes, sizeof(enable), cursor_evs);
-	errcode = errno;
-	fclose(cursor_evs);
-	evdi_log("Enabling cursor events on /dev/dri/card%d %s %s",
-		handle->device_index,
-		written < sizeof(enable) ? "failed: " : "succeeded",
-		written < sizeof(enable) ? strerror(errcode) : "");
+	do_ioctl(handle->fd, DRM_IOCTL_EVDI_ENABLE_CURSOR_EVENTS, &cmd, "enable cursor events");
 }
 
 void evdi_grab_pixels(evdi_handle handle,
@@ -824,15 +839,16 @@ static struct evdi_mode to_evdi_mode(struct drm_evdi_event_mode_changed *event)
 	return mode;
 }
 
-static uint64_t evdi_get_dumb_offset(evdi_handle ehandle, uint32_t handle)
+static int evdi_get_dumb_offset(evdi_handle ehandle, uint32_t handle, uint64_t *offset)
 {
 	struct drm_mode_map_dumb map_dumb = { 0 };
+	int ret = 0;
 
 	map_dumb.handle = handle;
-	do_ioctl(ehandle->fd, DRM_IOCTL_MODE_MAP_DUMB, &map_dumb,
+	ret = do_ioctl(ehandle->fd, DRM_IOCTL_MODE_MAP_DUMB, &map_dumb,
 		 "DRM_MODE_MAP_DUMB");
-
-	return map_dumb.offset;
+	*offset = map_dumb.offset;
+	return ret;
 }
 
 static struct evdi_cursor_set to_evdi_cursor_set(
@@ -845,15 +861,20 @@ static struct evdi_cursor_set to_evdi_cursor_set(
 	cursor_set.width =  event->width;
 	cursor_set.height = event->height;
 	cursor_set.enabled = event->enabled;
-	cursor_set.buffer_length = event->buffer_length;
+	cursor_set.buffer_length = 0;
 	cursor_set.buffer = NULL;
 	cursor_set.pixel_format = event->pixel_format;
 	cursor_set.stride = event->stride;
 
 	if (event->enabled) {
 		size_t size = event->buffer_length;
-		uint64_t offset =
-			evdi_get_dumb_offset(handle, event->buffer_handle);
+		uint64_t offset = 0;
+
+		if (evdi_get_dumb_offset(handle, event->buffer_handle, &offset)) {
+			evdi_log("Error: DRM_IOCTL_MODE_MAP_DUMB failed with error: %s", strerror(errno));
+			return cursor_set;
+		}
+
 		void *ptr = mmap(0, size, PROT_READ,
 				 MAP_SHARED, handle->fd, offset);
 
@@ -861,6 +882,9 @@ static struct evdi_cursor_set to_evdi_cursor_set(
 			cursor_set.buffer = malloc(size);
 			memcpy(cursor_set.buffer, ptr, size);
 			munmap(ptr, size);
+			cursor_set.buffer_length = size;
+		} else {
+			evdi_log("Error: mmap failed with error: %s", strerror(errno));
 		}
 	}
 
@@ -936,10 +960,16 @@ static void evdi_handle_event(evdi_handle handle,
 		if (evtctx->cursor_set_handler) {
 			struct drm_evdi_event_cursor_set *event =
 				(struct drm_evdi_event_cursor_set *) e;
+			struct evdi_cursor_set cursor_set = to_evdi_cursor_set(handle, event);
 
-			evtctx->cursor_set_handler(to_evdi_cursor_set(handle,
-								      event),
-						   evtctx->user_data);
+			if (cursor_set.enabled && cursor_set.buffer == NULL) {
+				evdi_log("Error: Cursor buffer is null!");
+				evdi_log("Disabling cursor events");
+				evdi_enable_cursor_events(handle, false);
+				cursor_set.enabled = false;
+				cursor_set.buffer_length = 0;
+			}
+			evtctx->cursor_set_handler(cursor_set, evtctx->user_data);
 		}
 		break;
 
